from __future__ import absolute_import

import time
import humanfriendly as hf

class Timer(object):
    """Prints the elapsed time after the execution of a block of code
       finishes.
    """
    def __init__(self, verbose=True):
        self._verbose = verbose
        self._tstart = None
        self.elapsed = None

    def __enter__(self):
        self._tstart = time.time()
        return self

    def __exit__(self, type_, value_, traceback_):
        self.elapsed = time.time() - self._tstart
        if self._verbose:
            print('Elapsed time: %s.' % hf.format_timespan(self.elapsed))



"""
.. module:: hdf5
   :synopsis: Querying hdf5 files.

.. moduleauthor:: Danilo Horta <horta@ebi.ac.uk>


"""

import h5py
import numpy as np
import os
import tempfile
import shutil
import asciitree


def fetch(fp, path):
    """Fetches an array from hdf5 file.

    :param str fp: hdf5 file path.
    :param str path: path inside the hdf5 file.
    :returns: An :class:`numpy.ndarray` representation of the corresponding hdf5 dataset.
    """
    with h5py.File(fp, 'r') as f:
        return f[path][:]

def tree(f_or_filepath, root_name='/', ret=False, show_chunks=False):
    """Shows a human-friendly tree representation of the contents of
    a hdf5 file.

    :param f_or_filepath: hdf5 file path or a reference to an open one.
    :param str root_name: group to be the root of the tree.
    :param bool ret: Whether to return a string or print it.
    :param bool show_chunks: show the chunks.
    :returns str: String representation if is `ret=True`.
    """
    if isinstance(f_or_filepath, str):
        with h5py.File(f_or_filepath, 'r') as f:
            return _tree(f, root_name, ret, show_chunks)
    else:
        return _tree(f_or_filepath, root_name, ret, show_chunks)

def _findnth(haystack, needle, n):
    parts= haystack.split(needle, n+1)
    if len(parts)<=n+1:
        return -1
    return len(haystack)-len(parts[-1])-len(needle)

def _visititems(root, func, level=0, prefix=''):
    if root.name != '/':
        name = root.name
        eman = name[::-1]
        i1 = _findnth(eman, '/', level)
        name = '/' + eman[:i1][::-1]
        func(prefix + name, root)
    if not hasattr(root, 'keys'):
        return
    for k in root.keys():
        if root.file == root[k].file:
            _visititems(root[k], func, level+1, prefix)
        else:
            _visititems(root[k], func, 0, prefix + root.name)

def _tree(f, root_name='/', ret=False, show_chunks=False):
    _names = []
    def get_names(name, obj):
        if isinstance(obj, h5py.Dataset):
            dtype = str(obj.dtype)
            shape = str(obj.shape)
            if show_chunks:
                chunks = str(obj.chunks)
                _names.append("%s [%s, %s, %s]" % (name[1:], dtype, shape,
                                                   chunks))
            else:
                _names.append("%s [%s, %s]" % (name[1:], dtype, shape))
        else:
            _names.append(name[1:])

    # f._visititems(get_names)
    _visititems(f, get_names)
    class Node(object):
        def __init__(self, name, children):
            self.name = name
            self.children = children

        def __str__(self):
            return self.name
    root = Node(root_name, dict())

    def add_to_node(node, ns):
        if len(ns) == 0:
            return
        if ns[0] not in node.children:
            node.children[ns[0]] = Node(ns[0], dict())
        add_to_node(node.children[ns[0]], ns[1:])

    _names = sorted(_names)
    for n in _names:
        ns = n.split('/')
        add_to_node(root, ns)

    def child_iter(node):
        keys = node.children.keys()
        indices = np.argsort(keys)
        indices = np.asarray(indices)
        return list(np.asarray(node.children.values())[indices])

    msg = asciitree.draw_tree(root, child_iter)
    if ret:
        return msg
    print(msg)

def copy_memmap_h5dt(arr, dt):
    """Copies a :class:`numpy.memmap` to a hdf5 dataset."""
    if arr.ndim > 2:
        raise Exception("I don't know how to handle arrays" +
                        " with more than 2 dimensions yet.")
    assert arr.shape == dt.shape
    if arr.ndim == 1:
        dt[:] = arr[:]
    else:
        if dt.chunks is not None:
            chunk_row = dt.chunks[0]
        else:
            chunk_row = 512
        r = 0
        while r < arr.shape[0]:
            re = r + chunk_row
            re = min(re, arr.shape[0])
            dt[r:re,:] = arr[r:re,:]
            r = re

def _good_chunk(dataset):
    if hasattr(dataset, 'chunks'):
        return dataset.chunks
    c = 256
    ndim = len(dataset.shape)
    chunks = (min(c, dataset.shape[i]) for i in range(ndim))
    return chunks

def change_layout(fp, path, chunks, compression='gzip'):
    """Changes layout of a HDF5 dataset."""
    import dask.array as da

    def do(f):
        dataset = f[path]
        gpath = os.path.dirname('/' + path)
        g = f[gpath]

        d = da.from_array(dataset, chunks=_good_chunk(dataset))

        name = os.path.basename(dataset.name)
        tmp_name = name + '_tmp_cfm92askj3'
        if tmp_name in g:
            del g[tmp_name]
        tmp_d = g.create_dataset(tmp_name, shape=dataset.shape,
                             dtype=dataset.dtype, chunks=chunks,
                             compression=compression)
        da.store(d, tmp_d)

        del g[name]
        g[name] = g[tmp_name]
        del g[tmp_name]

    if isinstance(fp, str):
        with h5py.File(fp, 'r+') as f:
            do(f)
    else:
        do(fp)

def change_layout_greedy(fp, path, chunks, compression='lzf', shuffle=True):
    """Changes layout of a HDF5 dataset, in a greedy manner."""
    def do(f):
        dataset = f[path]
        gpath = os.path.dirname('/' + path)
        g = f[gpath]

        d = dataset.value

        name = os.path.basename(dataset.name)
        tmp_name = name + '_tmp_cfm92askj3'
        if tmp_name in g:
            del g[tmp_name]

        g.create_dataset(tmp_name, data=d, shape=dataset.shape,
                         dtype=dataset.dtype, chunks=chunks,
                         compression=compression, shuffle=shuffle)

        del g[name]
        g[name] = g[tmp_name]
        del g[tmp_name]

    if isinstance(fp, str):
        with h5py.File(fp, 'r+') as f:
            do(f)
    else:
        do(fp)

def convert_matrices_to_row_layout(f):
    """Changes layout of a HDF5 2-by-2 dataset into C-matrix layout."""
    def foo(path, node, f):
        if isinstance(node, h5py.Dataset) and len(node.shape) == 2:
            change_layout_greedy(f, path, chunks=(1, node.shape[1]))
    _visititems(f, lambda path, node: foo(path, node, f))

def convert_matrices_to_col_layout(f):
    """Changes layout of a HDF5 2-by-2 dataset into Fortran-matrix layout."""
    def foo(path, node, f):
        if isinstance(node, h5py.Dataset) and len(node.shape) == 2:
            change_layout_greedy(f, path, chunks=(node.shape[0], 1))
    _visititems(f, lambda path, node: foo(path, node, f))

def copy_h5dt_memmap_filepath(dt, fp):
    """Copies a HDF5 dataset to a `numpy.memmap`."""

    arr = np.memmap(fp, mode='w+', shape=dt.shape, dtype=dt.dtype)
    if arr.ndim > 2:
        raise Exception("I don't know how to handle arrays" +
                        " with more than 2 dimensions yet.")
    assert arr.shape == dt.shape

    if len(dt.shape) == 1:
        arr[:] = dt[:]
        del arr
    else:
        if dt.chunks is not None:
            chunk_row = dt.chunks[0]
        else:
            chunk_row = 512
        r = 0
        del arr
        while r < dt.shape[0]:
            arr = np.memmap(fp, mode='r+', shape=dt.shape, dtype=dt.dtype)
            re = r + chunk_row
            re = min(re, dt.shape[0])
            s = np.s_[r:re,:]
            dt.read_direct(arr, s, s)
            r = re
            del arr

class Memmap(object):
    """Represents a HDF5 dataset as a `numpy.memmap`."""
    def __init__(self, filepath, path, readonly=True, tmp_folder=None):
        self._filepath = filepath
        self._path = path
        self._folder = None
        self._X = None
        self._readonly = readonly
        self._tmp_folder = tmp_folder

    def __enter__(self):
        self._folder = tempfile.mkdtemp(dir=self._tmp_folder)
        with h5py.File(self._filepath, 'r', libversion='latest') as f:
            dt = f[self._path]
            shape = dt.shape
            dtype = dt.dtype
            copy_h5dt_memmap_filepath(dt, os.path.join(self._folder, 'X'))

        mode = 'r' if self._readonly else 'r+'
        X = np.memmap(os.path.join(self._folder, 'X'), mode=mode,
                      shape=shape, dtype=dtype)
        self._X = X
        return X
    def __exit__(self, *args):
        del self._X
        shutil.rmtree(self._folder)

class XBuffRows(object):
    """Interates over HDF5 dataset rows, buffering beforehand."""
    def __init__(self, X, row_indices, col_slice, buff_size=1000):
        buff_size = min(min(buff_size, X.shape[0]), len(row_indices))
        self._X = X
        self._row = 0
        self._row_buff = -1
        self._buff_size = buff_size
        self._row_indices = row_indices
        self._col_slice = col_slice
        ncols = len(np.empty(X.shape[1])[col_slice])
        self._Xbuff = np.empty((len(row_indices), ncols), dtype=X.dtype)
        self._arr_buff = np.empty(buff_size, dtype=X.dtype)

    def __iter__(self):
        return self

    def _extract_buffer(self, row_indices):
        cs = self._col_slice
        if isinstance(self._X, h5py.Dataset):
            srii = np.argsort(row_indices)
            sri = row_indices[srii]
            try:
                self._X.read_direct(self._Xbuff, np.s_[sri,cs], np.s_[srii,:])
            except TypeError:
                for (i, csi) in enumerate(cs):
                    self._X.read_direct(self._arr_buff, np.s_[sri,csi],
                                        np.s_[:])
                    self._Xbuff[np.s_[srii,i]] = self._arr_buff
        else:
            self._Xbuff[:len(row_indices),:] = self._X[row_indices,cs]

    def next(self):
        if self._row >= len(self._row_indices):
            raise StopIteration

        if self._row_buff == -1:
            l = self._row
            r = l + self._buff_size
            r = min(r, len(self._row_indices))
            self._extract_buffer(self._row_indices[l:r])
            self._row_buff = 0

        vec = self._Xbuff[self._row_buff,:]
        self._row_buff += 1
        if self._row_buff >= self._Xbuff.shape[0]:
            self._row_buff = -1
        self._row += 1

        return vec

def do_convert_layout(args):
    with h5py.File(args.filepath, 'r+') as f:
        if args.type == 'row':
            convert_matrices_to_row_layout(f)
        elif args.type == 'col':
            convert_matrices_to_col_layout(f)
        else:
            raise ValueError('Unknown layout type: %s.' % args.type)

def do_see(args):
    tree(args.filepath, show_chunks=args.show_chunks)

def entry_point():
    from argparse import ArgumentParser
    p = ArgumentParser()
    sub = p.add_subparsers()

    s = sub.add_parser('see')
    s.add_argument('filepath')
    s.add_argument('--show-chunks', dest='show_chunks', action='store_true')
    s.add_argument('--no-show-chunks', dest='show_chunks', action='store_false')
    s.set_defaults(func=do_see, show_chunks=False)

    s = sub.add_parser('convert-layout')
    s.add_argument('filepath')
    s.add_argument('type')
    s.set_defaults(func=do_convert_layout)

    args = p.parse_args()
    func = args.func
    del args.func
    func(args)



from __future__ import absolute_import as _absolute_import

from pkg_resources import get_distribution as _get_distribution
from pkg_resources import DistributionNotFound as _DistributionNotFound

from . import dict
from . import hdf5
from . import inspect
from . import object
from . import path
from . import pickle
from . import report
from . import scalar
from . import set
from . import string
from . import sys
from . import time
from . import symbol

try:
    __version__ = _get_distribution('limix_util').version
except _DistributionNotFound:
    __version__ = 'unknown'


def test():
    import os
    p = __import__('limix_util').__path__[0]
    src_path = os.path.abspath(p)
    old_path = os.getcwd()
    os.chdir(src_path)

    try:
        return_code = __import__('pytest').main(['-q'])
    finally:
        os.chdir(old_path)

    if return_code == 0:
        print("Congratulations. All tests have passed!")

    return return_code



from __future__ import absolute_import
import progressbar as pbm
from time import time
import sys

class ProgressBar(object):
    """Wrapper around `progressbar.Progressbar` to make it easier to use."""
    def __init__(self, n):
        self._pb = pbm.ProgressBar(widgets=[pbm.Percentage(), pbm.Bar()],
                                   maxval=n).start()

    def update(self, i):
        self._pb.update(i)

    def finish(self):
        self._pb.finish()

class BeginEnd(object):
    """Prints message before and after the end of block of code."""
    def __init__(self, task, silent=False):
        self._task = str(task)
        self._start = None
        self._silent = silent

    def __enter__(self):
        self._start = time()
        if not self._silent:
            print('-- %s start --' % self._task)
            sys.stdout.flush()

    def __exit__(self, *args):
        elapsed = time() - self._start
        if not self._silent:
            print('-- %s end (%.2f s) --' % (self._task, elapsed))
            sys.stdout.flush()

def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")



def isfloat(value):
    """Tests whether the provided object can be converted to a float."""
    try:
        float(value)
        return True
    except ValueError:
        return False

def isint(value):
    """Tests whether the provided object can be converted to an integer without
    losing information."""
    try:
        return float(value) == int(value)
    except ValueError:
        return False

def isnumber(value):
    """Returns isfloat(value) or isint(value)."""
    return isfloat(value) or isint(value)



from __future__ import absolute_import
import gzip
try:
    import cPickle as pickle_
except ImportError:
    import pickle as pickle_
import os
import collections
from . import report
from . import object as object_
from . import path
from os.path import join
from os.path import isdir

class PickleByName(object):
    """Makes un-pickle-able objects pick-able by setting its un-pickle-able
    attributes as signature only attributes."""
    def __init__(self):
        self._signature_only_attrs = set()

    def set_signature_only_attr(self, attr_name):
        self._signature_only_attrs.add(attr_name)

    def __getstate__(self):
        d = self.__dict__.copy()
        for attr_name in self._signature_only_attrs:
            o = getattr(self, attr_name)

            d[attr_name + '_fullname'] = object_.fullname(o)
            d[attr_name + '_init_dict'] = o.init_dict()
            del d[attr_name]
        return d

    def __setstate__(self, d):
        import importlib

        for attr_name in d['_signature_only_attrs']:
            fn = d[attr_name + '_fullname']
            k = fn.rfind(".")
            module_name, class_name = fn[:k], fn[k+1:]
            init_dict = d[attr_name + '_init_dict']
            mod = importlib.import_module(module_name)
            class_ = getattr(mod, class_name)
            o = class_(**init_dict)
            d[attr_name] = o
            del d[attr_name + '_fullname']
            del d[attr_name + '_init_dict']

        self.__dict__.update(d)

class SlotPickleMixin(object):
    """Top-class that allows mixing of classes with and without slots.

    Takes care that instances can still be pickled with the lowest
    protocol. Moreover, provides a generic `__dir__` method that
    lists all slots.

    """

    # We want to allow weak references to the objects
    __slots__ = ['__weakref__']

    def _get_all_slots(self):
        """Returns all slots as set"""
        all_slots = (getattr(cls, '__slots__', [])
                         for cls in self.__class__.__mro__)
        return set(slot for slots in all_slots for slot in slots)

    def __getstate__(self):
        if hasattr(self, '__dict__'):
            # We don't require that all sub-classes also define slots,
            # so they may provide a dictionary
            statedict = self.__dict__.copy()
        else:
            statedict = {}
        # Get all slots of potential parent classes
        for slot in self._get_all_slots():
            try:
                value = getattr(self, slot)
                statedict[slot] = value
            except AttributeError:
                pass
        # Pop slots that cannot or should not be pickled
        statedict.pop('__dict__', None)
        statedict.pop('__weakref__', None)
        return statedict

    def __setstate__(self, state):
        for key, value in state.items():
            setattr(self, key, value)

    def __dir__(self):
        result = dir(self.__class__)
        result.extend(self._get_all_slots())
        if hasattr(self, '__dict__'):
            result.extend(self.__dict__.keys())
        return result

def pickle(obj, filepath):
    with gzip.open(filepath, 'wb', compresslevel=9) as f:
        pickle_.dump(obj, f, -1)

def unpickle(filepath):
    with gzip.open(filepath, 'rb', compresslevel=9) as f:
        return pickle_.load(f)

def _save_cache(folder, lastmodif_hash):
    fpath = join(folder, '.folder_hash')
    with open(fpath, 'w') as f:
        f.write(lastmodif_hash)

def _get_file_list(folder):
    file_list = []
    for (dir_, _, files) in os.walk(folder):
        if dir_ == folder:
            continue
        for f in files:
            fpath = join(dir_, f)
            if fpath.endswith('pkl') and os.path.basename(fpath) != 'all.pkl':
                file_list.append(fpath)
    return file_list

def _merge(file_list):
    pbar = report.ProgressBar(len(file_list))
    out = dict()
    for (i, fpath) in enumerate(file_list):
        d = unpickle(fpath)
        if isinstance(d, collections.Iterable):
            out.update(d)
        else:
            key = os.path.basename(fpath).split('.')[0]
            out[int(key)] = d
        pbar.update(i+1)
    pbar.finish()
    return out

def pickle_merge(folder):
    """Merges pickle files from the specified folder and save it to `all.pkl`.
    """
    file_list = _get_file_list(folder)

    if len(file_list) == 0:
        print('There is nothing to merge because no file'+
              ' has been found in %s.' % folder)
        return

    with report.BeginEnd('Computing hashes'):
        ha = path.folder_hash(folder, ['all.pkl', '.folder_hash'])

    subfolders = [d for d in os.listdir(folder) if isdir(join(folder, d))]

    with path.temp_folder() as tf:
        for sf in subfolders:
            path.make_sure_path_exists(join(tf, sf))
            path.cp(join(folder, sf), join(tf, sf))
        file_list = _get_file_list(tf)

        with report.BeginEnd('Merging pickles'):
            out = _merge(file_list)

    with report.BeginEnd('Storing pickles'):
        pickle(out, join(folder, 'all.pkl'))
    _save_cache(folder, ha)

    return out



import os
import sys
from setuptools import setup
from setuptools import find_packages


def setup_package():
    src_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    old_path = os.getcwd()
    os.chdir(src_path)
    sys.path.insert(0, src_path)

    needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
    pytest_runner = ['pytest-runner'] if needs_pytest else []

    setup_requires = [] + pytest_runner
    install_requires = ['pytest', 'scipy>=0.17', 'numpy>=1.9',
                        'tabulate', 'humanfriendly',
                        'progressbar2', 'asciitree', 'mock']
    tests_require = install_requires

    metadata = dict(
        name='limix_util',
        version='1.0.1',
        maintainer="Danilo Horta",
        maintainer_email="horta@ebi.ac.uk",
        license="MIT",
        url='http://github.com/Horta/limix-util',
        packages=find_packages(),
        zip_safe=True,
        install_requires=install_requires,
        setup_requires=setup_requires,
        tests_require=tests_require,
        include_package_data=True,
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Operating System :: OS Independent",
        ],
    )

    try:
        from distutils.command.bdist_conda import CondaDistribution
    except ImportError:
        pass
    else:
        metadata['distclass'] = CondaDistribution
        metadata['conda_buildnum'] = 1
        metadata['conda_features'] = ['mkl']

    try:
        setup(**metadata)
    finally:
        del sys.path[0]
        os.chdir(old_path)

if __name__ == '__main__':
    setup_package()



from __future__ import print_function
import contextlib
import errno
import tempfile
import shutil
import os
import subprocess
import hashlib
import sys
from distutils.spawn import find_executable


@contextlib.contextmanager
def temp_folder():
    folder = tempfile.mkdtemp()
    try:
        yield folder
    finally:
        shutil.rmtree(folder)

def count_lines(filepath):
    return sum(1 for line in open(filepath, 'r'))

class TmpFileCopy(object):
    """Creates a temporary file copy."""
    def __init__(self, path, copy_back=False):
        self._dst = None
        self._path = path
        self._folder = None
        self._copy_back = copy_back
        if not os.path.exists(path):
            raise Exception('%s does not exist.' % path)

    def __enter__(self):
        self._folder = tempfile.mkdtemp()
        bname = os.path.basename(self._path)
        dst = os.path.join(self._folder, bname)

        retcode = subprocess.call("cp " + self._path + " " + dst,
                                  shell=True)

        if retcode < 0:
            print("Child was terminated by signal %d" % retcode,
                  file=sys.stderr)
            raise Exception('Could not copy %s.' % self._path)

        self._dst = dst
        return dst

    def __exit__(self, *args):
        if self._copy_back:
            retcode = subprocess.call("cp " + self._dst + " " + self._path,
                                      shell=True)

            if retcode < 0:
                print("Child was terminated by signal %d" % retcode,
                      file=sys.stderr)
                raise Exception('Could not copy %s back.' % self._dst)
        shutil.rmtree(self._folder)

class ChDir(object):
    """Steps into a directory temporarily.
    """
    def __init__(self, path):
        self.old_dir = os.getcwd()
        self.new_dir = path

    def __enter__(self):
        os.chdir(self.new_dir)

    def __exit__(self, *args):
        os.chdir(self.old_dir)

def make_sure_path_exists(path):
    """Creates a path recursively if necessary.
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise

def cp(folder_src, folder_dst):
    """Uses cp shell command to copy files from a folder to another.
    """
    retcode = subprocess.call("cp " + folder_src + "/* " + folder_dst,
                              shell=True)

    if retcode < 0:
        print("Child was terminated by signal %d" % retcode,
              file=sys.stderr)

def rrm(paths):
    """Uses rm shell command to remove files.
    """
    paths = paths if isinstance(paths, list) else [paths]
    if len(paths) == 0:
        return

    cmd = subprocess.list2cmdline(['rm', '-rf'] + paths)
    retcode = subprocess.call(cmd, shell=True)

    if retcode < 0:
        print("Child was terminated by signal %d" % retcode,
              file=sys.stderr)

def touch(fname, times=None):
    """Creates an empty file.
    """
    with open(fname, 'a'):
        os.utime(fname, times)

def bin_exists(name):
    """Checks whether an executable file exists.
    """
    return find_executable(name) is not None

def folder_hash(folder, exclude_files=None):
    """Recursively compute the hash of all files in a folder and sum it up.
    """
    if exclude_files is None:
        exclude_files = []

    if not bin_exists('md5deep'):
        raise EnvironmentError("Couldn't not find md5deep.")

    out = subprocess.check_output('md5deep -r %s' % folder, shell=True)
    lines = sorted(out.strip('\n').split('\n'))

    m = hashlib.md5()
    for line in lines:
        hash_ = line[0:32]
        fp = line[34:]
        if os.path.basename(fp) not in exclude_files:
            m.update(hash_)
    return m.hexdigest()
